/* ARCADEmini USB driver for Windows
 * version 1.0.0 date 2004-08-25
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#ifndef INC_AmUsb
#define INC_AmUsb



//*** modes of the driver

//defines the behaviour when using normal file operations as read, write, seek
#define AmUsbModeNone -1 //do not change current mode
#define AmUsbModeNull 0 //act as null device
#define AmUsbModeStream 1 //display stream written into device
#define AmUsbModeCf 2 //act as compact flash reader / writer

//*** device dependent constants
#define AmUsbOutStyleLed 0x00 //normal LED matrix
#define AmUsbOutStyleLatch 0x01 //LATCHed circuit



//*** structures for ioctls

//set command pointer
struct tAmUsbSetCmdPtrIn
{
  unsigned long SecNo; //sector part of command pointer
  unsigned short Ofs; //offset part of command pointer
} stAmUsbSetCmdPtrIn;
struct tAmUsbSetCmdPtrOut
{
  unsigned long SecNo; //sector part of command pointer
  unsigned short Ofs; //offset part of command pointer
} stAmUsbSetCmdPtrOut;
//get command pointer
struct tAmUsbGetCmdPtrOut
{
  unsigned long SecNo; //sector part of command pointer
  unsigned short Ofs; //offset part of command pointer
} stAmUsbGetCmdPtrOut;
//load pixel buffer
struct tAmUsbLoadPixBufIn
{
  unsigned char Pixels[260]; //the pixel data read from the pixel buffer
} stAmUsbLoadPixBufIn;
//read pixel buffer
struct tAmUsbReadPixBufOut
{
  unsigned char Pixels[260]; //the pixel data to load into the pixel buffer
} stAmUsbReadPixBufOut;
//load overlay pixel buffer
struct tAmUsbLoadOvpBufIn
{
  unsigned char Pixels[260]; //the pixel data read from the overlay pixel buffer
} stAmUsbLoadOvpBufIn;
//read overlay pixel buffer
struct tAmUsbReadOvpBufOut
{
  unsigned char Pixels[260]; //the pixel data to load into the overlay pixel buffer
} stAmUsbReadOvpBufOut;
//load sector buffer
struct tAmUsbLoadSecBufIn
{
  unsigned char Data[512]; //the data to load into the sector buffer
} stAmUsbLoadSecBufIn;
//read sector buffer
struct tAmUsbReadSecBufOut
{
  unsigned char Data[512]; //the data read from the sector buffer
} stAmUsbReadSecBufOut;
//get CF sector count
struct tAmUsbGetCfSecCntOut
{
  unsigned long SecCnt; //number of sectors on compact flash
} stAmUsbGetCfSecCntOut;
//write CF sector
struct tAmUsbWriteCfIn
{
  unsigned long SecNo; //sector number on compact flash
} stAmUsbWriteCfIn;
struct tAmUsbWriteCfOut
{
  unsigned long SecNo; //sector number on compact flash
  unsigned char Err; //error code (0 for success)
} stAmUsbWriteCfOut;
//read CF sector
struct tAmUsbReadCfIn
{
  unsigned long SecNo; //sector number on compact flash
} stAmUsbReadCfIn;
struct tAmUsbReadCfOut
{
  unsigned long SecNo; //sector number on compact flash
  unsigned char Err; //error code (0 for success)
} stAmUsbReadCfOut;
//get internal sector count
struct tAmUsbGetIfSecCntOut
{
  unsigned long SecCnt; //number of sectors in internal flash
} stAmUsbGetIfSecCntOut;
//read internal sector
struct tAmUsbReadIfIn
{
  unsigned long SecNo; //sector number in internal flash
} stAmUsbReadIfIn;
struct tAmUsbReadIfOut
{
  unsigned long SecNo; //sector number in internal flash
  unsigned char Err; //error code (0 for success)
} stAmUsbReadIfOut;
//get compact flash flag
struct tAmUsbGetCfFlagOut
{
  unsigned char CfFlag; //!= 0 if commands are executed from compact flash
} stAmUsbGetCfFlagOut;
//load variables
struct tAmUsbLoadVarsIn
{
  unsigned char Vars[256]; //the contents of the variables
} stAmUsbLoadVarsIn;
//read variables
struct tAmUsbReadVarsOut
{
  unsigned char Vars[256]; //the contents of the variables
} stAmUsbReadVarsOut;
//load stack
struct tAmUsbLoadStackIn
{
  unsigned char Data[128]; //the contents of the stack
} stAmUsbLoadStackIn;
//read stack
struct tAmUsbReadStackOut
{
  unsigned char Data[128]; //the contents of the stack
} stAmUsbReadStackOut;
//set stack pointer
struct tAmUsbSetStackPtrIn
{
  unsigned char StackPtr; //the new value for the stack pointer (0-127)
} stAmUsbSetStackPtrIn;
struct tAmUsbSetStackPtrOut
{
  unsigned char StackPtr; //the new value for the stack pointer (0-127)
} stAmUsbSetStackPtrOut;
//get stack pointer
struct tAmUsbGetStackPtrOut
{
  unsigned char StackPtr; //the value of the stack pointer (0-127)
} stAmUsbGetStackPtrOut;
//get and set key state
struct tAmUsbGetSetKeysIn
{
  unsigned char Pressed, PressedMask; //how to modify pressed key state
  unsigned char Released, ReleasedMask; //how to modify released key state
} stAmUsbGetSetKeysIn;
struct tAmUsbGetSetKeysOut
{
  unsigned char State, Pressed, Released; //key state read
} stAmUsbGetSetKeysOut;
//get and set SNES0 key state
struct tAmUsbGetSetSnes0In
{
  unsigned short Pressed, PressedMask; //how to modify pressed key state
  unsigned short Released, ReleasedMask; //how to modify released key state
} stAmUsbGetSetSnes0In;
struct tAmUsbGetSetSnes0Out
{
  unsigned short State, Pressed, Released; //key state read
} stAmUsbGetSetSnes0Out;
//get and set SNES1 key state
struct tAmUsbGetSetSnes1In
{
  unsigned short Pressed, PressedMask; //how to modify pressed key state
  unsigned short Released, ReleasedMask; //how to modify released key state
} stAmUsbGetSetSnes1In;
struct tAmUsbGetSetSnes1Out
{
  unsigned short State, Pressed, Released; //key state read
} stAmUsbGetSetSnes1Out;
//set time
struct tAmUsbSetTimeIn
{
  unsigned char Sec, Min, Hour; //the time to set the ARCADEmini clock to
} stAmUsbSetTimeIn;
//get time
struct tAmUsbGetTimeOut
{
  unsigned char Sec, Min, Hour; //the time of the ARCADEmini clock
} stAmUsbGetTimeOut;
//get firmware version
struct tAmUsbGetFwVersionOut
{
  unsigned char Major, Minor, Revision; //version
  unsigned char Year, Month, Day; //release date of the firmware
} stAmUsbGetFwVersionOut;
//get ID
struct tAmUsbGetIdOut
{
  unsigned char Id[16]; //the ID
  unsigned char Err; //error code (0 for success)
} stAmUsbGetIdOut;
//set ID
struct tAmUsbSetIdIn
{
  unsigned char Id[16]; //the ID
} stAmUsbSetIdIn;
struct tAmUsbSetIdOut
{
  unsigned char Err; //error code (0 for success)
} stAmUsbSetIdOut;
//get output style
struct tAmUsbGetOutStyleOut
{
  unsigned char Style; //the output style (0x00=LED, 0x01=LATCH)
  unsigned char Err; //error code (0 for success)
} stAmUsbGetOutStyleOut;
//set output style
struct tAmUsbSetOutStyleIn
{
  unsigned char Style; //the output style (0x00=LED, 0x01=LATCH)
} stAmUsbSetOutStyleIn;
struct tAmUsbSetOutStyleOut
{
  unsigned char Err; //error code (0 for success)
} stAmUsbSetOutStyleOut;
//get output timing
struct tAmUsbGetOutTimingOut
{
  unsigned char Timing[15]; //the output timing (0..50)
  unsigned char Err; //error code (0 for success)
} stAmUsbGetOutTimingOut;
//set output timing
struct tAmUsbSetOutTimingIn
{
  unsigned char Timing[15]; //the output timing (0..50)
} stAmUsbSetOutTimingIn;
struct tAmUsbSetOutTimingOut
{
  unsigned char Err; //error code (0 for success)
} stAmUsbSetOutTimingOut;
//get version information
struct tAmUsbGetVersionOut
{
  unsigned char Major, Minor, Revision; //version
  unsigned short Year; //release date
  unsigned char Month, Day;
} stAmUsbGetVersionOut;
//set mode
struct tAmUsbSetModeIn
{
  int Mode; //the new mode
  int DefMode; //the new default mode
} stAmUsbSetModeIn;
struct tAmUsbSetModeOut
{
  int Mode; //the new mode
  int DefMode; //the new default mode
} stAmUsbSetModeOut;
//get mode
struct tAmUsbGetModeOut
{
  int Mode; //the current mode
  int DefMode; //the current default mode
} stAmUsbGetModeOut;



//*** ioctls

//needed stuff from devioctl.h
#ifndef CTL_CODE
  #define CTL_CODE( DeviceType, Function, Method, Access ) (((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method))
#endif
#ifndef FILE_DEVICE_UNKNOWN
  #define FILE_DEVICE_UNKNOWN 0x00000022
#endif
#ifndef METHOD_BUFFERED
  #define METHOD_BUFFERED 0
#endif
#ifndef FILE_ANY_ACCESS
  #define FILE_ANY_ACCESS 0
#endif
//halt ARCADEmini
//turns off command processing
#define AmUsbIoctlNoHalt CTL_CODE( FILE_DEVICE_UNKNOWN, 0x00, METHOD_BUFFERED, FILE_ANY_ACCESS )
//run ARCADEmini
//turns on command processing
#define AmUsbIoctlNoRun CTL_CODE( FILE_DEVICE_UNKNOWN, 0x01, METHOD_BUFFERED, FILE_ANY_ACCESS )
//step ARCADEmini
//processes a single command
#define AmUsbIoctlNoStep CTL_CODE( FILE_DEVICE_UNKNOWN, 0x02, METHOD_BUFFERED, FILE_ANY_ACCESS )
//set command pointer
#define AmUsbIoctlNoSetCmdPtr CTL_CODE( FILE_DEVICE_UNKNOWN, 0x08, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get command pointer
#define AmUsbIoctlNoGetCmdPtr CTL_CODE( FILE_DEVICE_UNKNOWN, 0x09, METHOD_BUFFERED, FILE_ANY_ACCESS )
//load pixel buffer
//loads the pixel buffer with 260 bytes of data
#define AmUsbIoctlNoLoadPixBuf CTL_CODE( FILE_DEVICE_UNKNOWN, 0x10, METHOD_BUFFERED, FILE_ANY_ACCESS )
//read pixel buffer
//reads 260 bytes of data from the pixel buffer
#define AmUsbIoctlNoReadPixBuf CTL_CODE( FILE_DEVICE_UNKNOWN, 0x11, METHOD_BUFFERED, FILE_ANY_ACCESS )
//load overlay pixel buffer
//loads the overlay pixel buffer with 260 bytes of data
#define AmUsbIoctlNoLoadOvpBuf CTL_CODE( FILE_DEVICE_UNKNOWN, 0x12, METHOD_BUFFERED, FILE_ANY_ACCESS )
//read overlay pixel buffer
//reads 260 bytes of data from the overlay pixel buffer
#define AmUsbIoctlNoReadOvpBuf CTL_CODE( FILE_DEVICE_UNKNOWN, 0x13, METHOD_BUFFERED, FILE_ANY_ACCESS )
//swap display buffers
//exchanges the foreground and background pixel buffer
#define AmUsbIoctlNoSwapBufs CTL_CODE( FILE_DEVICE_UNKNOWN, 0x20, METHOD_BUFFERED, FILE_ANY_ACCESS )
//swap display buffers and copy image
//exchanges the foreground and background pixel buffer
//then copies the foreground picture into the background buffer
#define AmUsbIoctlNoSwapBufsCp CTL_CODE( FILE_DEVICE_UNKNOWN, 0x21, METHOD_BUFFERED, FILE_ANY_ACCESS )
//load sector buffer
//loads the sector buffer with 512 bytes of data
#define AmUsbIoctlNoLoadSecBuf CTL_CODE( FILE_DEVICE_UNKNOWN, 0x30, METHOD_BUFFERED, FILE_ANY_ACCESS )
//read sector buffer
//reads 512 bytes of data from the sector buffer
#define AmUsbIoctlNoReadSecBuf CTL_CODE( FILE_DEVICE_UNKNOWN, 0x31, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get CF sector count
//returns the number of sectors on compact flash
#define AmUsbIoctlNoGetCfSecCnt CTL_CODE( FILE_DEVICE_UNKNOWN, 0x40, METHOD_BUFFERED, FILE_ANY_ACCESS )
//write CF sector
//writes the sector buffer to a sector on compact flash
#define AmUsbIoctlNoWriteCf CTL_CODE( FILE_DEVICE_UNKNOWN, 0x42, METHOD_BUFFERED, FILE_ANY_ACCESS )
//read CF sector
//reads a sector on compact flash into sector buffer
#define AmUsbIoctlNoReadCf CTL_CODE( FILE_DEVICE_UNKNOWN, 0x43, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get internal sector count
//returns the number of sectors in internal flash
#define AmUsbIoctlNoGetIfSecCnt CTL_CODE( FILE_DEVICE_UNKNOWN, 0x44, METHOD_BUFFERED, FILE_ANY_ACCESS )
//read internal sector
//reads a sector in internal flash into sector buffer
#define AmUsbIoctlNoReadIf CTL_CODE( FILE_DEVICE_UNKNOWN, 0x47, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get compact flash flag
//returns if commands are executed from compact flash or from internal flash
#define AmUsbIoctlNoGetCfFlag CTL_CODE( FILE_DEVICE_UNKNOWN, 0x48, METHOD_BUFFERED, FILE_ANY_ACCESS )
//load variables
//loads the variables with 256 bytes of data
#define AmUsbIoctlNoLoadVars CTL_CODE( FILE_DEVICE_UNKNOWN, 0x50, METHOD_BUFFERED, FILE_ANY_ACCESS )
//read variables
//reads 256 bytes of data from the variables
#define AmUsbIoctlNoReadVars CTL_CODE( FILE_DEVICE_UNKNOWN, 0x51, METHOD_BUFFERED, FILE_ANY_ACCESS )
//load stack
//loads the stack with 128 bytes of data
#define AmUsbIoctlNoLoadStack CTL_CODE( FILE_DEVICE_UNKNOWN, 0x54, METHOD_BUFFERED, FILE_ANY_ACCESS )
//read stack
//reads 128 bytes of data from the stack
#define AmUsbIoctlNoReadStack CTL_CODE( FILE_DEVICE_UNKNOWN, 0x55, METHOD_BUFFERED, FILE_ANY_ACCESS )
//set stack pointer
#define AmUsbIoctlNoSetStackPtr CTL_CODE( FILE_DEVICE_UNKNOWN, 0x56, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get stack pointer
#define AmUsbIoctlNoGetStackPtr CTL_CODE( FILE_DEVICE_UNKNOWN, 0x57, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get and set key state
//gets key state and sets key state afterwards
#define AmUsbIoctlNoGetSetKeys CTL_CODE( FILE_DEVICE_UNKNOWN, 0x60, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get and set SNES0 key state
//gets SNES0 key state and sets SNES0 key state afterwards
#define AmUsbIoctlNoGetSetSnes0 CTL_CODE( FILE_DEVICE_UNKNOWN, 0x62, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get and set SNES1 key state
//gets SNES1 key state and sets SNES1 key state afterwards
#define AmUsbIoctlNoGetSetSnes1 CTL_CODE( FILE_DEVICE_UNKNOWN, 0x63, METHOD_BUFFERED, FILE_ANY_ACCESS )
//set time
//sets the time of the ARCADEmini closk
#define AmUsbIoctlNoSetTime CTL_CODE( FILE_DEVICE_UNKNOWN, 0x70, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get time
//gets the time from the ARCADEmini clock
#define AmUsbIoctlNoGetTime CTL_CODE( FILE_DEVICE_UNKNOWN, 0x71, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get firmware version
//returns the firmware version of ARCADEmini
#define AmUsbIoctlNoGetFwVersion CTL_CODE( FILE_DEVICE_UNKNOWN, 0x80, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get ID
//gets the ID of this ARCADEmini
#define AmUsbIoctlNoGetId CTL_CODE( FILE_DEVICE_UNKNOWN, 0x84, METHOD_BUFFERED, FILE_ANY_ACCESS )
//set ID
//sets the ID of this ARCADEmini
#define AmUsbIoctlNoSetId CTL_CODE( FILE_DEVICE_UNKNOWN, 0x85, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get output style
//gets the output style
#define AmUsbIoctlNoGetOutStyle CTL_CODE( FILE_DEVICE_UNKNOWN, 0x88, METHOD_BUFFERED, FILE_ANY_ACCESS )
//set output style
//sets the output style
#define AmUsbIoctlNoSetOutStyle CTL_CODE( FILE_DEVICE_UNKNOWN, 0x89, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get output timing
//gets the output timing
#define AmUsbIoctlNoGetOutTiming CTL_CODE( FILE_DEVICE_UNKNOWN, 0x8A, METHOD_BUFFERED, FILE_ANY_ACCESS )
//set output timing
//sets the output timing
#define AmUsbIoctlNoSetOutTiming CTL_CODE( FILE_DEVICE_UNKNOWN, 0x8B, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get version information
//returns version information about the driver
#define AmUsbIoctlNoGetVersion CTL_CODE( FILE_DEVICE_UNKNOWN, 0xF0, METHOD_BUFFERED, FILE_ANY_ACCESS )
//set mode
//sets the driver mode for normal file operations
#define AmUsbIoctlNoSetMode CTL_CODE( FILE_DEVICE_UNKNOWN, 0xF2, METHOD_BUFFERED, FILE_ANY_ACCESS )
//get mode
//gets the driver mode for normal file operations
#define AmUsbIoctlNoGetMode CTL_CODE( FILE_DEVICE_UNKNOWN, 0xF3, METHOD_BUFFERED, FILE_ANY_ACCESS )



#endif