/* ARCADEmini USB driver for Windows test application
 * version 1.0 date 2004-08-25
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <windows.h>
#include <stdio.h>

#include "AmUsb.h"



#define count( array ) (sizeof( (array) ) / sizeof( (array)[0] ))



char * SNESKEYS[16] = { "?0", "?1", "?2", "?3", "R", "L", "X", "A", "Right", "Left", "Down", "Up", "Start", "Select", "Y", "B" };



//get first word and rest of a string
static void StrFirstWord( char * pString, char * * ppFirst, char * * ppRest )
{
	//jump over whitespaces
	while( *pString == '\t' || *pString == ' ' )
		pString++;

	//first word starts here
	*ppFirst = pString;

	//jump over first word
	while( *pString != 0 && *pString != '\t' && *pString != ' ' )
		pString++;

	//first word ends here
	if( *pString != 0 )
	{
		*pString = 0;
		pString++;
	}

	//jump over whitespaces
	while( *pString == '\t' || *pString == ' ' )
		pString++;

	//rest starts here
	*ppRest = pString;
}

//convert error number into string
char ErrStrBuffer[1024];
static char * ErrNo2Str( unsigned long Err )
{
	char * pChr;
	FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, NULL, Err, 0, ErrStrBuffer, sizeof( ErrStrBuffer ), NULL );
	if( pChr = strchr( ErrStrBuffer, '\r' ) )
		*pChr = 0;
	if( pChr = strchr( ErrStrBuffer, '\n' ) )
		*pChr = 0;
	return ErrStrBuffer;
}

//check if commands are executed from compact flash or internal flash
static void CmdCfFlag( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetCfFlagOut GetCfFlagOut;
	unsigned long Len;
	int FuRes;
	
	//get compact flash flag
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetCfFlag, NULL, 0, &GetCfFlagOut, sizeof( GetCfFlagOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show if commands are executed from compact flash or internal flash
	else
		if( GetCfFlagOut.CfFlag == 0 )
			printf( "  commands are executed from internal flash\n" );
		else
			printf( "  commands are executed from compact flash\n" );
}

//read compact flash sector into sector buffer
static void CmdCfRead( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	struct tAmUsbReadCfIn ReadCfIn;
	struct tAmUsbReadCfOut ReadCfOut;
	unsigned long Len;
	int FuRes;

	//get sector number
	StrFirstWord( pParams, &pParam, &pParams );
	ReadCfIn.SecNo = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 )
	{
		printf( "  invalid sector number: \"%s\"\n", pParam );
		return;
	}
		 
	//read sector of compact flash
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadCf, &ReadCfIn, sizeof( ReadCfIn ), &ReadCfOut, sizeof( ReadCfOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show sector number and result
	else
		if( ReadCfOut.Err == 0 )
			printf( "  compact flash sector %ld successfully read\n", ReadCfOut.SecNo );
		else
			printf( "  reading compact flash sector %ld failed\n"
							"   - no CF inserted?\n"
							"   - reading past end of CF?\n", ReadCfOut.SecNo );
}

//get number of sectors on compact flash
static void CmdCfSecs( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetCfSecCntOut GetCfSecCntOut;
	unsigned long Len;
	int FuRes;

	//get number of sectors on compact flash
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetCfSecCnt, NULL, 0, &GetCfSecCntOut, sizeof( GetCfSecCntOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show number of sectors and size
	else
		printf( "  compact flash has got %ld sectors (%f MB)\n",
						GetCfSecCntOut.SecCnt,
						(double)GetCfSecCntOut.SecCnt / 2048.0 );
}

//write sector buffer to compact flash sector
static void CmdCfWrite( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	struct tAmUsbWriteCfIn WriteCfIn;
	struct tAmUsbWriteCfOut WriteCfOut;
	unsigned long Len;
	int FuRes;

	//get sector number
	StrFirstWord( pParams, &pParam, &pParams );
	WriteCfIn.SecNo = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 )
	{
		printf( "  invalid sector number: \"%s\"\n", pParam );
		return;
	}
		 
	//write sector of compact flash
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoWriteCf, &WriteCfIn, sizeof( WriteCfIn ), &WriteCfOut, sizeof( WriteCfOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show sector number and result
	else
		if( WriteCfOut.Err == 0 )
			printf( "  compact flash sector %ld successfully written\n", WriteCfOut.SecNo );
		else
			printf( "  writing compact flash sector %ld failed\n"
							"   - no CF inserted?\n"
							"   - writing past end of CF?\n", WriteCfOut.SecNo );
}

//get command pointer
static void CmdCmdPtrGet( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetCmdPtrOut GetCmdPtrOut;
	unsigned long Len;
	int FuRes;

	//get command pointer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetCmdPtr, NULL, 0, &GetCmdPtrOut, sizeof( GetCmdPtrOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show command pointer
	else
		printf( "  command pointer: sector %ld offset %X\n",
						GetCmdPtrOut.SecNo, GetCmdPtrOut.Ofs );
}

//set command pointer
static void CmdCmdPtrSet( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbSetCmdPtrIn SetCmdPtrIn;
	struct tAmUsbSetCmdPtrOut SetCmdPtrOut;
	char * pParam, * pChr;
	unsigned long Val, Len;
	int FuRes;

	//get sector number
	StrFirstWord( pParams, &pParam, &pParams );
	Val = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 )
	{
		printf( "  invalid sector number: \"%s\"\n", pParam );
		return;
	}
	SetCmdPtrIn.SecNo = Val;

	//get offset
	StrFirstWord( pParams, &pParam, &pParams );
	Val = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 || Val > 511 )
	{
		printf( "  invalid offset: \"%s\"\n", pParam );
		return;
	}
	SetCmdPtrIn.Ofs = (unsigned short)Val;

	//set command pointer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoSetCmdPtr, &SetCmdPtrIn, sizeof( SetCmdPtrIn ), &SetCmdPtrOut, sizeof( SetCmdPtrOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show command pointer
	else
		printf( "  command pointer: sector %ld offset %X\n",
						SetCmdPtrOut.SecNo, SetCmdPtrOut.Ofs );
}

//get version of firmware
static void CmdFwVer( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetFwVersionOut GetFwVersionOut;
	unsigned long Len;
	int FuRes;

	//get firmware version
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetFwVersion, NULL, 0, &GetFwVersionOut, sizeof( GetFwVersionOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show firmware version
	else
		printf( "  firmware version %d.%d.%d date %04d-%02d-%02d\n",
						GetFwVersionOut.Major, GetFwVersionOut.Minor, GetFwVersionOut.Revision,
						2000 + GetFwVersionOut.Year, GetFwVersionOut.Month, GetFwVersionOut.Day );
}

//halt ARCADEmini
static void CmdHalt( HANDLE hAmUsb, char * pParams )
{
	int FuRes;
	unsigned long Len;

	//halt ARCADEmini
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoHalt, NULL, 0, NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
}

//get ID of this ARCADEmini
static void CmdIdGet( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetIdOut GetIdOut;
	int FuRes, I;
	unsigned long Len;

	//get identification
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetId, NULL, 0, &GetIdOut, sizeof( GetIdOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show identification
	else
	{
		//success
		if( GetIdOut.Err == 0 )
		{
			printf( "  identification: " );
			for( I = 0; I < sizeof( GetIdOut.Id ); I++ )
				if( (char)GetIdOut.Id[I] >= ' ' )
					printf( "%c", (char)GetIdOut.Id[I] );
				else
					printf( " " );
			printf( "\n" );
		}
		//error
		else
			printf( "  reading identification failed (maybe EEPROM is busy)\n" );
	}
}

//set ID of this ARCADEmini
static void CmdIdSet( HANDLE hAmUsb, char * pParams )
{
	char * pParam;
	struct tAmUsbSetIdIn SetIdIn;
	struct tAmUsbSetIdOut SetIdOut;
	int FuRes, I;
	unsigned long Len;

	//get identification from parameter
	StrFirstWord( pParams, &pParam, &pParams );
	for( I = 0; pParam[I] != 0 && I < sizeof( SetIdIn.Id ); I++ )
		SetIdIn.Id[I] = (unsigned char)pParam[I];
	for( ; I < sizeof( SetIdIn.Id ); I++ )
		SetIdIn.Id[I] = (unsigned char)' ';

	//set identification
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoSetId, &SetIdIn, sizeof( SetIdIn ), &SetIdOut, sizeof( SetIdOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show result
	else
	{
		//error
		if( SetIdOut.Err != 0 )
			printf( "  writing identification failed (maybe EEPROM is busy)\n" );
	}
}

//read internal flash sector into sector buffer
static void CmdIfRead( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	struct tAmUsbReadIfIn ReadIfIn;
	struct tAmUsbReadIfOut ReadIfOut;
	unsigned long Len;
	int FuRes;

	//get sector number
	StrFirstWord( pParams, &pParam, &pParams );
	ReadIfIn.SecNo = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 )
	{
		printf( "  invalid sector number: \"%s\"\n", pParam );
		return;
	}
		 
	//read sector of internal flash
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadIf, &ReadIfIn, sizeof( ReadIfIn ), &ReadIfOut, sizeof( ReadIfOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show sector number and result
	else
		if( ReadIfOut.Err == 0 )
			printf( "  internal flash sector %ld successfully read\n", ReadIfOut.SecNo );
		else
			printf( "  reading internal flash sector %ld failed\n"
							"   - reading past end of internal flash?\n", ReadIfOut.SecNo );
}

//get number of sectors in internal flash
static void CmdIfSecs( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetIfSecCntOut GetIfSecCntOut;
	unsigned long Len;
	int FuRes;

	//get number of sectors in internal flash
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetIfSecCnt, NULL, 0, &GetIfSecCntOut, sizeof( GetIfSecCntOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show number of sectors and size
	else
		printf( "  internal flash has got %ld sectors (%f KB)\n",
						GetIfSecCntOut.SecCnt,
						(double)GetIfSecCntOut.SecCnt / 2 );
}

//get key state (and clear it)
static void CmdKeys( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetSetKeysIn GetSetKeysIn;
	struct tAmUsbGetSetKeysOut GetSetKeysOut;
	unsigned long Len;
	int FuRes, I;
	unsigned char Mask;

	//clear key state
	GetSetKeysIn.Pressed = 0x00;
	GetSetKeysIn.PressedMask = 0xFF;
	GetSetKeysIn.Released = 0x00;
	GetSetKeysIn.ReleasedMask = 0xFF;

	//get and set key state
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetSetKeys, &GetSetKeysIn, sizeof( GetSetKeysIn ), &GetSetKeysOut, sizeof( GetSetKeysOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show key state
	else
	{
		printf( "  state:" );
		for( I = 0, Mask = 1; I < 8; I++, Mask <<= 1 )
			if( GetSetKeysOut.State & Mask )
				printf( " %d", I + 1 );
		printf( "\n" );
		printf( "  pressed:" );
		for( I = 0, Mask = 1; I < 8; I++, Mask <<= 1 )
			if( GetSetKeysOut.Pressed & Mask )
				printf( " %d", I + 1 );
		printf( "\n" );
		printf( "  released:" );
		for( I = 0, Mask = 1; I < 8; I++, Mask <<= 1 )
			if( GetSetKeysOut.Released & Mask )
				printf( " %d", I + 1 );
		printf( "\n" );
	}
}

//get current mode and default mode
static void CmdModeGet( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetModeOut GetModeOut;
	unsigned long Len;
	int FuRes;

	//get mode and default mode
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetMode, NULL, 0, &GetModeOut, sizeof( GetModeOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show mode and default mode
	else
	{
		//show mode
		switch( GetModeOut.Mode )
		{
			case AmUsbModeNull: printf( "  mode: null\n" ); break;
			case AmUsbModeStream: printf( "  mode: stream\n" ); break;
			case AmUsbModeCf: printf( "  mode: cf\n" ); break;
			default: printf( "  mode: <unknown>\n" );
		}
		//show default mode
		switch( GetModeOut.DefMode )
		{
			case AmUsbModeNull: printf( "  default mode: null\n" ); break;
			case AmUsbModeStream: printf( "  default mode: stream\n" ); break;
			case AmUsbModeCf: printf( "  default mode: cf\n" ); break;
			default: printf( "  default mode: <unknown>\n" );
		}
	}
}

//set current mode and default mode
static void CmdModeSet( HANDLE hAmUsb, char * pParams )
{
	char * pMode, * pDefMode;
	struct tAmUsbSetModeIn SetModeIn;
	struct tAmUsbSetModeOut SetModeOut;
	unsigned long Len;
	int FuRes;

	//get mode
	StrFirstWord( pParams, &pMode, &pParams );
	if( pMode[0] == 0 )
		SetModeIn.Mode = AmUsbModeNone;
	else if( stricmp( pMode, "null" ) == 0 )
		SetModeIn.Mode = AmUsbModeNull;
	else if( stricmp( pMode, "stream" ) == 0 )
		SetModeIn.Mode = AmUsbModeStream;
	else if( stricmp( pMode, "cf" ) == 0 )
		SetModeIn.Mode = AmUsbModeCf;
	else
	{
		printf( "  invalid mode: \"%s\"\n", pMode );
		return;
	}
	
	//get default mode
	StrFirstWord( pParams, &pDefMode, &pParams );
	if( pDefMode[0] == 0 )
		SetModeIn.DefMode = AmUsbModeNone;
	else if( stricmp( pDefMode, "null" ) == 0 )
		SetModeIn.DefMode = AmUsbModeNull;
	else if( stricmp( pDefMode, "stream" ) == 0 )
		SetModeIn.DefMode = AmUsbModeStream;
	else if( stricmp( pDefMode, "cf" ) == 0 )
		SetModeIn.DefMode = AmUsbModeCf;
	else
	{
		printf( "  invalid mode: \"%s\"\n", pDefMode );
		return;
	}

	//get mode and default mode
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoSetMode, &SetModeIn, sizeof( SetModeIn ), &SetModeOut, sizeof( SetModeOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show mode and default mode
	else
	{
		//show mode
		switch( SetModeOut.Mode )
		{
			case AmUsbModeNull: printf( "  mode: null\n" ); break;
			case AmUsbModeStream: printf( "  mode: stream\n" ); break;
			case AmUsbModeCf: printf( "  mode: cf\n" ); break;
			default: printf( "  mode: <unknown>\n" );
		}
		//show default mode
		switch( SetModeOut.DefMode )
		{
			case AmUsbModeNull: printf( "  default mode: null\n" ); break;
			case AmUsbModeStream: printf( "  default mode: stream\n" ); break;
			case AmUsbModeCf: printf( "  default mode: cf\n" ); break;
			default: printf( "  default mode: <unknown>\n" );
		}
	}
}

//get output style
static void CmdOutStyleGet( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetOutStyleOut GetOutStyleOut;
	unsigned long Len;
	int FuRes;

	//get output style
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetOutStyle, NULL, 0, &GetOutStyleOut, sizeof( GetOutStyleOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show output style
	else
	{
		//success
		if( GetOutStyleOut.Err == 0 )
		{
			printf( "  output style: " );
			switch( GetOutStyleOut.Style )
			{
				case AmUsbOutStyleLed: printf( "LED" ); break;
				case AmUsbOutStyleLatch: printf( "LATCH" ); break;
				default: printf( "unknown (0x%02X)", GetOutStyleOut.Style );
			};
			printf( "\n" );
		}
		//error
		else
			printf( "  reading output style failed (maybe EEPROM is busy)\n" );
	}
}

//set output style to LED or LATCH
static void CmdOutStyleSet( HANDLE hAmUsb, char * pParams )
{
	char * pParam;
	struct tAmUsbSetOutStyleIn SetOutStyleIn;
	struct tAmUsbSetOutStyleOut SetOutStyleOut;
	unsigned long Len;
	int FuRes;

	//get output style from parameter
	StrFirstWord( pParams, &pParam, &pParams );
	if( stricmp( pParam, "led" ) == 0 )
		SetOutStyleIn.Style = AmUsbOutStyleLed;
	else if( stricmp( pParam, "latch" ) == 0 )
		SetOutStyleIn.Style = AmUsbOutStyleLatch;
	else
	{
		printf( "  invalid output style: \"%s\"\n", pParam );
		return;
	}

	//set output style
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoSetOutStyle, &SetOutStyleIn, sizeof( SetOutStyleIn ), &SetOutStyleOut, sizeof( SetOutStyleOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show result
	else
	{
		//error
		if( SetOutStyleOut.Err != 0 )
			printf( "  writing output style failed (maybe EEPROM is busy)\n" );
	}
}

//get output timing
static void CmdOutTimingGet( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetOutTimingOut GetOutTimingOut;
	unsigned long Len;
	int FuRes, I;

	//get output timing
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetOutTiming, NULL, 0, &GetOutTimingOut, sizeof( GetOutTimingOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show output timing
	else
	{
		//success
		if( GetOutTimingOut.Err == 0 )
		{
			printf( "  grayscale:      1  2  3  4  5  6  7  8  9 10 11 12 13 14 15\n" );
			printf( "  output timing:" );
			for( I = 0; I < sizeof( GetOutTimingOut.Timing ); I++ )
				printf( " %2d", GetOutTimingOut.Timing[I] );
			printf( "\n" );
		}
		//error
		else
			printf( "  reading output timing failed (maybe EEPROM is busy)\n" );
	}
}

//set output timing
static void CmdOutTimingSet( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	struct tAmUsbSetOutTimingIn SetOutTimingIn;
	struct tAmUsbSetOutTimingOut SetOutTimingOut;
	unsigned long Len, Value;
	int FuRes, I;

	//get output timing from parameters
	for( I = 0; I < count( SetOutTimingIn.Timing ); I++ )
	{
		StrFirstWord( pParams, &pParam, &pParams );
		Value = strtoul( pParam, &pChr, 0 );
		if( pParam[0] == 0 || pChr[0] != 0 || Value > 50 )
		{
			printf( "  invalid timing for grayscale %d: \"%s\"\n", I + 1, pParam );
			return;
		}
		SetOutTimingIn.Timing[I] = (unsigned char)Value;
	}

	//set output timing
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoSetOutTiming, &SetOutTimingIn, sizeof( SetOutTimingIn ), &SetOutTimingOut, sizeof( SetOutTimingOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show result
	else
	{
		//error
		if( SetOutTimingOut.Err != 0 )
			printf( "  writing output timing failed (maybe EEPROM is busy)\n" );
	}
}

//load (a part of) the overlay pixel buffer
static void CmdOvpBufLoad( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	unsigned long Start, Value;
	struct tAmUsbReadOvpBufOut ReadOvpBufOut;
	struct tAmUsbLoadOvpBufIn LoadOvpBufIn;
	unsigned long Len;
	int FuRes, I, J;

	//get start address
	StrFirstWord( pParams, &pParam, &pParams );
	Start = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 || Start > 259 )
	{
		printf( "  invalid start address: \"%s\"\n", pParam );
		return;
	}

	//read overlay pixel buffer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadOvpBuf, NULL, 0, &ReadOvpBufOut, sizeof( ReadOvpBufOut ), &Len, NULL );

	//show error
	if( ! FuRes )
	{
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
		return;
	}

	//modify data
	memcpy( LoadOvpBufIn.Pixels, ReadOvpBufOut.Pixels, sizeof( ReadOvpBufOut.Pixels ) );
	for( I = (int)Start; I < 260; )
	{
		//get next data byte
		StrFirstWord( pParams, &pParam, &pParams );
		if( pParam[0] == 0 )
			break;
		Value = strtoul( pParam, &pChr, 0 );
		if( pParam[0] == 0 || pChr[0] != 0 || Value > 255 )
			printf( "  ignoring invalid data byte: \"%s\"\n", pParam );
		else
			LoadOvpBufIn.Pixels[I++] = (unsigned char)Value;
	}

	//load overlay pixel buffer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoLoadOvpBuf, &LoadOvpBufIn, sizeof( LoadOvpBufIn ), NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show overlay pixel buffer data
	else
	{
		for( I = 0; I < 20; I++ )
		{
			printf( "  %03d:", I * 13 );
			for( J = 0; J < 13; J++ )
				printf( " %02X", LoadOvpBufIn.Pixels[I * 13 + J] );
			printf( "\n" );
		}
	}
}

//read the overlay pixel buffer
static void CmdOvpBufRead( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbReadOvpBufOut ReadOvpBufOut;
	unsigned long Len;
	int FuRes, I, J;

	//read overlay pixel buffer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadOvpBuf, NULL, 0, &ReadOvpBufOut, sizeof( ReadOvpBufOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show overlay pixel buffer data
	else
	{
		for( I = 0; I < 20; I++ )
		{
			printf( "  %03d:", I * 13 );
			for( J = 0; J < 13; J++ )
				printf( " %02X", ReadOvpBufOut.Pixels[I * 13 + J] );
			printf( "\n" );
		}
	}
}

//load (a part of) the pixel buffer
static void CmdPixBufLoad( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	unsigned long Start, Value;
	struct tAmUsbReadPixBufOut ReadPixBufOut;
	struct tAmUsbLoadPixBufIn LoadPixBufIn;
	unsigned long Len;
	int FuRes, I, J;

	//get start address
	StrFirstWord( pParams, &pParam, &pParams );
	Start = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 || Start > 259 )
	{
		printf( "  invalid start address: \"%s\"\n", pParam );
		return;
	}

	//read pixel buffer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadPixBuf, NULL, 0, &ReadPixBufOut, sizeof( ReadPixBufOut ), &Len, NULL );

	//show error
	if( ! FuRes )
	{
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
		return;
	}

	//modify data
	memcpy( LoadPixBufIn.Pixels, ReadPixBufOut.Pixels, sizeof( ReadPixBufOut.Pixels ) );
	for( I = (int)Start; I < 260; )
	{
		//get next data byte
		StrFirstWord( pParams, &pParam, &pParams );
		if( pParam[0] == 0 )
			break;
		Value = strtoul( pParam, &pChr, 0 );
		if( pParam[0] == 0 || pChr[0] != 0 || Value > 255 )
			printf( "  ignoring invalid data byte: \"%s\"\n", pParam );
		else
			LoadPixBufIn.Pixels[I++] = (unsigned char)Value;
	}

	//load pixel buffer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoLoadPixBuf, &LoadPixBufIn, sizeof( LoadPixBufIn ), NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show pixel buffer data
	else
	{
		for( I = 0; I < 20; I++ )
		{
			printf( "  %03d:", I * 13 );
			for( J = 0; J < 13; J++ )
				printf( " %02X", LoadPixBufIn.Pixels[I * 13 + J] );
			printf( "\n" );
		}
	}
}

//read the pixel buffer
static void CmdPixBufRead( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbReadPixBufOut ReadPixBufOut;
	unsigned long Len;
	int FuRes, I, J;

	//read pixel buffer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadPixBuf, NULL, 0, &ReadPixBufOut, sizeof( ReadPixBufOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show pixel buffer data
	else
	{
		for( I = 0; I < 20; I++ )
		{
			printf( "  %03d:", I * 13 );
			for( J = 0; J < 13; J++ )
				printf( " %02X", ReadPixBufOut.Pixels[I * 13 + J] );
			printf( "\n" );
		}
	}
}

//run ARCADEmini
static void CmdRun( HANDLE hAmUsb, char * pParams )
{
	unsigned long Len;
	int FuRes;

	//run ARCADEmini
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoRun, NULL, 0, NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
}

//load (a part of) the sector buffer
static void CmdSecBufLoad( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	unsigned long Start, Value;
	struct tAmUsbReadSecBufOut ReadSecBufOut;
	struct tAmUsbLoadSecBufIn LoadSecBufIn;
	unsigned long Len;
	int FuRes, I, J;

	//get start address
	StrFirstWord( pParams, &pParam, &pParams );
	Start = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 || Start > 511 )
	{
		printf( "  invalid start address: \"%s\"\n", pParam );
		return;
	}

	//read sector buffer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadSecBuf, NULL, 0, &ReadSecBufOut, sizeof( ReadSecBufOut ), &Len, NULL );

	//show error
	if( ! FuRes )
	{
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
		return;
	}

	//modify data
	memcpy( LoadSecBufIn.Data, ReadSecBufOut.Data, sizeof( ReadSecBufOut.Data ) );
	for( I = (int)Start; I < 512; )
	{
		//get next data byte
		StrFirstWord( pParams, &pParam, &pParams );
		if( pParam[0] == 0 )
			break;
		Value = strtoul( pParam, &pChr, 0 );
		if( pParam[0] == 0 || pChr[0] != 0 || Value > 255 )
			printf( "  ignoring invalid data byte: \"%s\"\n", pParam );
		else
			LoadSecBufIn.Data[I++] = (unsigned char)Value;
	}

	//load sector buffer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoLoadSecBuf, &LoadSecBufIn, sizeof( LoadSecBufIn ), NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show sector buffer data
	else
	{
		for( I = 0; I < 32; I++ )
		{
			printf( "  %03X:", I << 4 );
			for( J = 0; J < 16; J++ )
				printf( " %02X", LoadSecBufIn.Data[(I << 4) + J] );
			printf( "\n" );
		}
	}
}

//read the sector buffer
static void CmdSecBufRead( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbReadSecBufOut ReadSecBufOut;
	unsigned long Len;
	int FuRes, I, J;

	//read sector buffer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadSecBuf, NULL, 0, &ReadSecBufOut, sizeof( ReadSecBufOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show sector buffer data
	else
	{
		for( I = 0; I < 32; I++ )
		{
			printf( "  %03X:", I << 4 );
			for( J = 0; J < 16; J++ )
				printf( " %02X", ReadSecBufOut.Data[(I << 4) + J] );
			printf( "\n" );
		}
	}
}

//get key SNES0 State (and clear it)
static void CmdSnes0( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetSetSnes0In GetSetSnes0In;
	struct tAmUsbGetSetSnes0Out GetSetSnes0Out;
	unsigned long Len;
	int FuRes, I;
	unsigned short Mask;

	//clear key state
	GetSetSnes0In.Pressed = 0x0000;
	GetSetSnes0In.PressedMask = 0xFFFF;
	GetSetSnes0In.Released = 0x0000;
	GetSetSnes0In.ReleasedMask = 0xFFFF;

	//get and set key state
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetSetSnes0, &GetSetSnes0In, sizeof( GetSetSnes0In ), &GetSetSnes0Out, sizeof( GetSetSnes0Out ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show key state
	else
	{
		printf( "  state:" );
		for( I = 0, Mask = 1; I < 16; I++, Mask <<= 1 )
			if( GetSetSnes0Out.State & Mask )
				printf( " %s", SNESKEYS[I] );
		printf( "\n" );
		printf( "  pressed:" );
		for( I = 0, Mask = 1; I < 16; I++, Mask <<= 1 )
			if( GetSetSnes0Out.Pressed & Mask )
				printf( " %s", SNESKEYS[I] );
		printf( "\n" );
		printf( "  released:" );
		for( I = 0, Mask = 1; I < 16; I++, Mask <<= 1 )
			if( GetSetSnes0Out.Released & Mask )
				printf( " %s", SNESKEYS[I] );
		printf( "\n" );
	}
}

//get key SNES1 State (and clear it)
static void CmdSnes1( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetSetSnes1In GetSetSnes1In;
	struct tAmUsbGetSetSnes1Out GetSetSnes1Out;
	unsigned long Len;
	int FuRes, I;
	unsigned short Mask;

	//clear key state
	GetSetSnes1In.Pressed = 0x0000;
	GetSetSnes1In.PressedMask = 0xFFFF;
	GetSetSnes1In.Released = 0x0000;
	GetSetSnes1In.ReleasedMask = 0xFFFF;

	//get and set key state
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetSetSnes1, &GetSetSnes1In, sizeof( GetSetSnes1In ), &GetSetSnes1Out, sizeof( GetSetSnes1Out ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show key state
	else
	{
		printf( "  state:" );
		for( I = 0, Mask = 1; I < 16; I++, Mask <<= 1 )
			if( GetSetSnes1Out.State & Mask )
				printf( " %s", SNESKEYS[I] );
		printf( "\n" );
		printf( "  pressed:" );
		for( I = 0, Mask = 1; I < 16; I++, Mask <<= 1 )
			if( GetSetSnes1Out.Pressed & Mask )
				printf( " %s", SNESKEYS[I] );
		printf( "\n" );
		printf( "  released:" );
		for( I = 0, Mask = 1; I < 16; I++, Mask <<= 1 )
			if( GetSetSnes1Out.Released & Mask )
				printf( " %s", SNESKEYS[I] );
		printf( "\n" );
	}
}

//load (a part of) the stack
static void CmdStackLoad( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	unsigned long Start, Value;
	struct tAmUsbReadStackOut ReadStackOut;
	struct tAmUsbLoadStackIn LoadStackIn;
	unsigned long Len;
	int FuRes, I, J;

	//get start address
	StrFirstWord( pParams, &pParam, &pParams );
	Start = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 || Start > 127 )
	{
		printf( "  invalid start address: \"%s\"\n", pParam );
		return;
	}

	//read stack
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadStack, NULL, 0, &ReadStackOut, sizeof( ReadStackOut ), &Len, NULL );

	//show error
	if( ! FuRes )
	{
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
		return;
	}

	//modify data
	memcpy( LoadStackIn.Data, ReadStackOut.Data, sizeof( ReadStackOut.Data ) );
	for( I = (int)Start; I < 128; )
	{
		//get next data byte
		StrFirstWord( pParams, &pParam, &pParams );
		if( pParam[0] == 0 )
			break;
		Value = strtoul( pParam, &pChr, 0 );
		if( pParam[0] == 0 || pChr[0] != 0 || Value > 255 )
			printf( "  ignoring invalid data byte: \"%s\"\n", pParam );
		else
			LoadStackIn.Data[I++] = (unsigned char)Value;
	}

	//load stack
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoLoadStack, &LoadStackIn, sizeof( LoadStackIn ), NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show sector buffer data
	else
	{
		for( I = 0; I < 8; I++ )
		{
			printf( "  %03X:", I << 4 );
			for( J = 0; J < 16; J++ )
				printf( " %02X", LoadStackIn.Data[(I << 4) + J] );
			printf( "\n" );
		}
	}
}

//read the stack
static void CmdStackRead( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbReadStackOut ReadStackOut;
	unsigned long Len;
	int FuRes, I, J;

	//read stack
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadStack, NULL, 0, &ReadStackOut, sizeof( ReadStackOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show sector buffer data
	else
	{
		for( I = 0; I < 8; I++ )
		{
			printf( "  %03X:", I << 4 );
			for( J = 0; J < 16; J++ )
				printf( " %02X", ReadStackOut.Data[(I << 4) + J] );
			printf( "\n" );
		}
	}
}

//get stack pointer
static void CmdStackPtrGet( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetStackPtrOut GetStackPtrOut;
	unsigned long Len;
	int FuRes;

	//get stack pointer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetStackPtr, NULL, 0, &GetStackPtrOut, sizeof( GetStackPtrOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show command pointer
	else
		printf( "  stack pointer: %X\n",
						GetStackPtrOut.StackPtr );
}

//set stack pointer
static void CmdStackPtrSet( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbSetStackPtrIn SetStackPtrIn;
	struct tAmUsbSetStackPtrOut SetStackPtrOut;
	char * pParam, * pChr;
	unsigned long Val, Len;
	int FuRes;

	//get stack pointer
	StrFirstWord( pParams, &pParam, &pParams );
	Val = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 || Val > 127 )
	{
		printf( "  invalid stack pointer Value: \"%s\"\n", pParam );
		return;
	}
	SetStackPtrIn.StackPtr = (unsigned char)Val;

	//set stack pointer
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoSetStackPtr, &SetStackPtrIn, sizeof( SetStackPtrIn ), &SetStackPtrOut, sizeof( SetStackPtrOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show command pointer
	else
		printf( "  stack pointer: %X\n",
						SetStackPtrOut.StackPtr );
}

//step ARCADEmini
static void CmdStep( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	unsigned long Cnt;
	unsigned long Len;
	int I, FuRes;

	//get number of steps
	StrFirstWord( pParams, &pParam, &pParams );
	Cnt = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 )
		Cnt = 1;
	else
	{ 
		if( pChr[0] != 0 || Cnt > 1000 )
		{
			printf( "  invalid step count: \"%s\"\n", pParam );
			return;
		}
	}

	//step ARCADEmini cnt times
	for( I = 0, FuRes = 1; I < (int)Cnt && FuRes; I++ )
		FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoStep, NULL, 0, NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error (after %d steps): %s (%d)\n", I, ErrNo2Str( GetLastError( ) ), GetLastError( ) );
}

//swap pixel buffers
static void CmdSwapBufs( HANDLE hAmUsb, char * pParams )
{
	unsigned long Len;
	int FuRes;

	//swap pixel buffers
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoSwapBufs, NULL, 0, NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
}

//swap pixel buffers and copy foreground to background buffer
static void CmdSwapBufsCp( HANDLE hAmUsb, char * pParams )
{
	unsigned long Len;
	int FuRes;

	//swap pixel buffers and copy image
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoSwapBufsCp, NULL, 0, NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
}

//get time from the ARCADEmini clock
static void CmdTimeGet( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetTimeOut GetTimeOut;
	unsigned long Len;
	int FuRes;

	//get time
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetTime, NULL, 0, &GetTimeOut, sizeof( GetTimeOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show time
	else
		printf( "  time is %02d:%02d:%02d\n",
						GetTimeOut.Hour, GetTimeOut.Min, GetTimeOut.Sec );
}

//set time of the ARCADEmini clock
static void CmdTimeSet( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	unsigned long Val;
	SYSTEMTIME LocalTime;
	struct tAmUsbSetTimeIn SetTimeIn;
	unsigned long Len;
	int FuRes;

	//time specified
	StrFirstWord( pParams, &pParam, &pParams );
	if( pParam[0] != 0 )
	{
		//get hour
		Val = strtoul( pParam, &pChr, 0 );
		if( pChr[0] != 0 || Val > 23 )
		{
			printf( "  invalid hour: \"%s\"\n", pParam );
			return;
		}
		SetTimeIn.Hour = (unsigned char)Val;
		//get minute
		StrFirstWord( pParams, &pParam, &pParams );
		Val = strtoul( pParam, &pChr, 0 );
		if( pParam[0] == 0 || pChr[0] != 0 || Val > 59 )
		{
			printf( "  invalid minute: \"%s\"\n", pParam );
			return;
		}
		SetTimeIn.Min = (unsigned char)Val;
		//get second
		StrFirstWord( pParams, &pParam, &pParams );
		Val = strtoul( pParam, &pChr, 0 );
		if( pParam[0] == 0 || pChr[0] != 0 || Val > 59 )
		{
			printf( "  invalid second: \"%s\"\n", pParam );
			return;
		}
		SetTimeIn.Sec = (unsigned char)Val;
	}
	//time not specified
	else
	{
		//get local time from system
		GetLocalTime( &LocalTime );
		SetTimeIn.Hour = (unsigned char)LocalTime.wHour;
		SetTimeIn.Min = (unsigned char)LocalTime.wMinute;
		SetTimeIn.Sec = (unsigned char)LocalTime.wSecond;
	}

	//set time
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoSetTime, &SetTimeIn, sizeof( SetTimeIn ), NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
}

//load (a part of) the variables
static void CmdVarsLoad( HANDLE hAmUsb, char * pParams )
{
	char * pParam, * pChr;
	unsigned long Start, Value;
	struct tAmUsbReadVarsOut ReadVarsOut;
	struct tAmUsbLoadVarsIn LoadVarsIn;
	unsigned long Len;
	int FuRes, I, J;

	//get start address
	StrFirstWord( pParams, &pParam, &pParams );
	Start = strtoul( pParam, &pChr, 0 );
	if( pParam[0] == 0 || pChr[0] != 0 || Start > 255 )
	{
		printf( "  invalid start address: \"%s\"\n", pParam );
		return;
	}

	//read variables
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadVars, NULL, 0, &ReadVarsOut, sizeof( ReadVarsOut ), &Len, NULL );

	//show error
	if( ! FuRes )
	{
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
		return;
	}

	//modify data
	memcpy( LoadVarsIn.Vars, ReadVarsOut.Vars, sizeof( ReadVarsOut.Vars ) );
	for( I = (int)Start; I < 256; )
	{
		//get next data byte
		StrFirstWord( pParams, &pParam, &pParams );
		if( pParam[0] == 0 )
			break;
		Value = strtoul( pParam, &pChr, 0 );
		if( pParam[0] == 0 || pChr[0] != 0 || Value > 255 )
			printf( "  ignoring invalid data byte: \"%s\"\n", pParam );
		else
			LoadVarsIn.Vars[I++] = (unsigned char)Value;
	}

	//load variables
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoLoadVars, &LoadVarsIn, sizeof( LoadVarsIn ), NULL, 0, &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show variables
	else
	{
		for( I = 0; I < 16; I++ )
		{
			printf( "  %03X:", I << 4 );
			for( J = 0; J < 16; J++ )
				printf( " %02X", LoadVarsIn.Vars[(I << 4) + J] );
			printf( "\n" );
		}
	}
}

//read the variables
static void CmdVarsRead( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbReadVarsOut ReadVarsOut;
	unsigned long Len;
	int FuRes, I, J;

	//read variables
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoReadVars, NULL, 0, &ReadVarsOut, sizeof( ReadVarsOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show variables
	else
	{
		for( I = 0; I < 16; I++ )
		{
			printf( "  %03X:", I << 4 );
			for( J = 0; J < 16; J++ )
				printf( " %02X", ReadVarsOut.Vars[(I << 4) + J] );
			printf( "\n" );
		}
	}
}

//get version of driver
static void CmdVer( HANDLE hAmUsb, char * pParams )
{
	struct tAmUsbGetVersionOut GetVersionOut;
	unsigned long Len;
	int FuRes;

	//get driver version
	FuRes = DeviceIoControl( hAmUsb, AmUsbIoctlNoGetVersion, NULL, 0, &GetVersionOut, sizeof( GetVersionOut ), &Len, NULL );

	//show error
	if( ! FuRes )
		printf( "  error: %s (%d)\n", ErrNo2Str( GetLastError( ) ), GetLastError( ) );
	//show driver version
	else
		printf( "  version %d.%d.%d, date %04d-%02d-%02d\n",
						GetVersionOut.Major, GetVersionOut.Minor, GetVersionOut.Revision,
						GetVersionOut.Year, GetVersionOut.Month, GetVersionOut.Day );
}

//help command prototype
static void CmdHelp( HANDLE hAmUsb, char * pParams );

//command table
struct tCmdTab
{
	char * Name;
	char * Parameters;
	char * Description;
	void (* Function)( HANDLE hAmUsb, char * pParams );
} CmdTab[] =
{
	{ "cf-flag", "", "check if commands are executed from compact flash or internal flash", CmdCfFlag },
	{ "cf-read", " <sector_no>", "read compact flash sector into sector buffer", CmdCfRead },
	{ "cf-secs", "", "get number of sectors on compact flash", CmdCfSecs },
	{ "cf-write", " <sector_no>", "write sector buffer to compact flash sector", CmdCfWrite },
	{ "cmd-ptr-get", "", "get command pointer", CmdCmdPtrGet },
	{ "cmd-ptr-set", " <sector_no> <offset>", "set command pointer", CmdCmdPtrSet },
	{ "end", "", "end the program", NULL },
	{ "exit", "", "end the program", NULL },
	{ "fw-ver", "", "show firmware version", CmdFwVer },
	{ "halt", "", "halt ARCADEmini", CmdHalt },
	{ "help", "", "you just found out what this does :-)", CmdHelp },
	{ "id-get", "", "get ID of this ARCADEmini", CmdIdGet },
	{ "id-set", " <id>", "set ID of this ARCADEmini", CmdIdSet },
	{ "if-read", " <sector_no>", "read internal flash sector into sector buffer", CmdIfRead },
	{ "if-secs", "", "get number of sectors in internal flash", CmdIfSecs },
	{ "keys", "", "get key state (and clear it)", CmdKeys },
	{ "mode-get", "", "get current mode and default mode", CmdModeGet },
	{ "mode-set", " [<mode> [<default mode>]]", "set current mode and default mode", CmdModeSet },
	{ "out-style-get", "", "get output style", CmdOutStyleGet },
	{ "out-style-set", " [led|latch]", "set output style to LED or LATCH", CmdOutStyleSet },
	{ "out-timing-get", "", "get output timing", CmdOutTimingGet },
	{ "out-timing-set", " <timing 1> ... <timing 15>", "set output timing", CmdOutTimingSet },
	{ "ovp-buf-load", " <start-addr> [<data> [...]]", "load (a part of) the overlay pixel buffer", CmdOvpBufLoad },
	{ "ovp-buf-read", "", "read the overlay pixel buffer", CmdOvpBufRead },
	{ "pix-buf-load", " <start-addr> [<data> [...]]", "load (a part of) the pixel buffer", CmdPixBufLoad },
	{ "pix-buf-read", "", "read the pixel buffer", CmdPixBufRead },
	{ "quit", "", "end the program", NULL },
	{ "run", "", "run ARCADEmini", CmdRun },
	{ "sec-buf-load", " <start-addr> [<data> [...]]", "load (a part of) the sector buffer", CmdSecBufLoad },
	{ "sec-buf-read", "", "read the sector buffer", CmdSecBufRead },
	{ "snes0", "", "get key SNES0 State (and clear it)", CmdSnes0 },
	{ "snes1", "", "get key SNES1 State (and clear it)", CmdSnes1 },
	{ "stack-load", " <start-addr> [<data> [...]]", "load (a part of) the stack", CmdStackLoad },
	{ "stack-read", "", "read the stack", CmdStackRead },
	{ "stack-ptr-get", "", "get stack pointer", CmdStackPtrGet },
	{ "stack-ptr-set", " <stack_ptr>", "set stack pointer", CmdStackPtrSet },
	{ "step", " [<count>]", "step ARCADEmini", CmdStep },
	{ "swap-bufs", "", "swap pixel buffers", CmdSwapBufs },
	{ "swap-bufs-cp", "", "swap pixel buffers and copy foreground to background buffer", CmdSwapBufsCp },
	{ "time-get", "", "get time from the ARCADEmini clock", CmdTimeGet },
	{ "time-set", " [<hour> <min> <sec>]", "set time of the ARCADEmini clock", CmdTimeSet },
	{ "vars-load", " <start-addr> [<data> [...]]", "load (a part of) the variables", CmdVarsLoad },
	{ "vars-read", "", "read the variables", CmdVarsRead },
	{ "ver", "", "show driver version", CmdVer },
};

//help command
static void CmdHelp( HANDLE hAmUsb, char * pParams )
{
	char * pCmd;
	int I;

	//get command
	StrFirstWord( pParams, &pCmd, &pParams );

	//command found in parameters
	if( pCmd[0] != 0 )
	{
		//search command
		for( I = 0; I < count( CmdTab ); I++ )
			if( stricmp( pCmd, CmdTab[I].Name ) == 0 )
				break;
		//command found in table
		if( I < count( CmdTab ) )
			//print help for this command
			printf( "  %s%s\n    %s\n", CmdTab[I].Name, CmdTab[I].Parameters, CmdTab[I].Description );
		//command not found in table
		else
			//print error message
			printf( "  no help found to \"%s\"\n", pCmd );
		return;
	}

	//print command table and descriptions
	for( I = 0; I < count( CmdTab ); I++ )
		printf( "  %s%s\n    %s\n", CmdTab[I].Name, CmdTab[I].Parameters, CmdTab[I].Description );
}

//main program
int main( int ArgCnt, char * * Args )
{
	HANDLE hAmUsb;
	int I;
	char Line[1024], * pChr, * pCmd, * pParams;

	//print message
	printf( "ARCADEmini USB driver for Windows test application\n"
					"version 1.0 date 2004-08-25\n"
					"part of the ARCADEmini project - http[s]://arcademini.schuermans.info/\n"
					"Copyright (C) 2003-2004 1stein <1stein@schuermans.info>\n"
					"Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html\n" );

	//check parameters
	if( ArgCnt < 2 )
	{
		printf( "  usage: %s <AmUsb device>\n", Args[0] );
		return -1;
	}

	//open AmUsb device
	hAmUsb = CreateFile( Args[1], GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
	if( hAmUsb == INVALID_HANDLE_VALUE )
	{
		printf( "  could not open AmUsb device \"%s\": %s\n", Args[1], ErrNo2Str( GetLastError( ) ) );
		return -1;
	}

	//main loop
	for( ; ; )
	{
		//read a line
		printf( "AmUsb> " ); //prompt
		fflush( stdout );
		Line[0] = 0; //read line
		fgets( Line, sizeof( Line ), stdin );
		pChr = strchr( Line, '\n' ); //remove newline
		if( pChr != NULL )
			*pChr = 0;

		//get command (first word of line)
		StrFirstWord( Line, &pCmd, &pParams );

		//ignore empty command
		if( pCmd[0] != 0 )
		{
			//find command in command table
			for( I = 0; I < count( CmdTab ); I++ )
				if( stricmp( pCmd, CmdTab[I].Name ) == 0 )
					break;
			//command found
			if( I < count( CmdTab ) )
			{
				//exit main loop if no function
				if( CmdTab[I].Function == NULL )
					break;
				//call command
				CmdTab[I].Function( hAmUsb, pParams );
			}
			//command not found
			else
				printf( "  unknown command - type \"help\"\n" );
		}
	}

	//close AmUsb device
	CloseHandle( hAmUsb );

	return 0;
}

